import pathlib
import shutil

from .cancel_sim import cancel_sim
from .clear_sim import clear_sim
from .._config import ORCH_IMAGE_NAME_TEMPLATE
from ..util.execute import execute
from ..util.config_data import ConfigData


def delete_sim_setup( setup_dir ):
    '''
    Delete a simulation setup, including all associated Docker images and containers.

    :param setup_dir: path to simulation setup (string)
    :return: return dict with status of build process:
        {
            'valid': flag indicating if deletion succeded (boolean)
            'status': detailed status message (string)
        }
    '''

    try:
        # Cancel all running simulations.
        cancel_sim( setup_dir, 'all' )

    except Exception as err:
        return dict( 
            valid = False, 
            status = 'cancelling running simulations failed:\n{}'.format( err )
        )

    try:
        # Clear all finished simulations.
        clear_sim( setup_dir, 'all' )

    except Exception as err:
        return dict( 
            valid = False, 
            status = 'clearing finished simulations failed:\n{}'.format( err )
        )


    try:
        # Retrieve simulation setup configuration.
        config_data = ConfigData( setup_dir )
        sim_setup_id = config_data['id'].strip()

        # Specify name of mosaik orchestrator image.
        orch_image_name = ORCH_IMAGE_NAME_TEMPLATE.format( sim_setup_id.lower() )
    
        execute( [
            'docker', 'image', 'rm', # Remove Docker image.
            orch_image_name # Specify image name.
        ] )

    except Exception as err:
        return dict( 
            valid = False, 
            status = 'remove orchestrator image failed:\n{}'.format( err )
        )

    try:
        setup_dir_path = pathlib.Path( setup_dir ).resolve()
        shutil.rmtree( setup_dir_path )
        
    except Exception as err:
        return dict( 
            valid = False, 
            status = 'deleting simulation setup failed:\n{}'.format( err )
        )

    return dict( 
        valid = True, 
        status = 'deletion of simulation setup succeeded: {}'.format( setup_dir_path ) 
    )


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Delete a simulation setup, including all associated Docker images and containers.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    args = parser.parse_args()

    try:
        delete_status = delete_sim_setup( args.setup_dir )

        print( delete_status['status'] )
        if True == delete_status['valid']:
            sys.exit( 0 )
        else:
            sys.exit( 1 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )


import pathlib

from ..util.config_data import ConfigData
from ..util.execute import execute, execute_and_capture_output


def get_sim_results( setup_dir, id, overwrite = False ):
    '''
    Retrieve the results of finished simulations.

    :param setup_dir: path to simulation setup (string)
    :param id: either 'all' or ID of finished simulation container (string)
    :param overwrite: overwrite previously retrieved results (boolean, default: False)
    :return: on success, return ID(s) of simulation(s) for which results have been retrieved (string)
    '''

    if not isinstance( overwrite, bool ):
        raise TypeError( 'Parameter \'overwrite\' must be of type \'bool\'' )
        
    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )
    sim_ids_down = config_data['sim_ids_down']
    config_data_orch = config_data['orchestrator']
    results = config_data_orch['results']

    if 'all' == id.lower():
        cp_ids = sim_ids_down
    else:
        cp_ids = [ id ]
        if id not in sim_ids_down:
            raise RuntimeError( 'No finished simulation (status \'DOWN\') with ID \'{}\''.format( id ) )


    for cp_id in cp_ids:
    
        # Retrieve working directory of Docker container.
        sim_working_dir = execute_and_capture_output( [
            'docker', 'inspect', # Return low-level information on Docker objects.
            '--format={{.Config.WorkingDir}}', # Only return the container's working directory.
            cp_id # Specify simulation ID (container name).
        ] )
    
        # Create local results folder, where all results are copied to.
        results_dir_path = pathlib.Path( setup_dir, cp_id ).resolve( strict = False )
        try:
            results_dir_path.mkdir( exist_ok = False )
        except FileExistsError:
            if ( False == overwrite ):
                raise RuntimeWarning( 'Results directory already exists: {}'.format( results_dir_path ) )
    
        # Copy each result file or folder from the container to the local results folder.
        for res in results:
            if pathlib.Path( res ).is_absolute():
                res_path = '{}:{}'.format( cp_id, res )
            else:
                res_path = '{}:{}/{}'.format( cp_id, sim_working_dir, res )
    
            execute( [
                'docker', 'cp', # Copy files/folders between a container and the local filesystem.
                res_path, # Specify container name (simulation ID) and file/folder to be copied.
                str( results_dir_path ) # Specify destination folder.
                ] )

    # On success, return the simulation ID(s).
    return cp_ids


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Retrieve the results of finished simulations.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    group = parser.add_mutually_exclusive_group( required = True )

    group.add_argument(
        '--id', 
        action = 'store',
        metavar = 'ID',
        help = 'simulation ID (Docker container name)'
    )

    group.add_argument(
        '--all',
        action = 'store_true',
        help = 'retrieve results from all finished simulation containers'
    )

    parser.add_argument(
        '--overwrite',
        action = 'store_true',
        help = 'overwrite previously retrieved results'
    )

    args = parser.parse_args()

    try:
        id = 'all' if ( True == args.all ) else args.id
        sim_ids = get_sim_results( args.setup_dir, id, args.overwrite )

        print( 'Retrieved results for simulation(s) with ID = {}'.format( ' '.join( sim_ids ) ) )
        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



import subprocess


def execute( cmd ):
    '''
    Invoke and execute a subprocess.

    :param cmd: sequence of program arguments or else a single string or path-like object (see Python "subprocess" module).
    :return: None
    '''
    p = subprocess.Popen(
        cmd,
        #stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
    )

    # Wait for process to terminate and retrieve return code.
    return_code = p.wait()

    if ( 0 != return_code ):
        raise Exception( p.stderr.read().decode( 'utf-8' )[:-1] )


def execute_and_capture_output( cmd ):
    '''
    Invoke and execute a subprocess.
    Output from the subprocess to stdout is captured and returned as string.

    :param cmd: sequence of program arguments or else a single string or path-like object (see Python "subprocess" module).
    :return: subprocess output to stdout (str)
    '''
    p = subprocess.Popen(
        cmd,
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
    )

    # Wait for process to terminate and retrieve return code.
    return_code = p.wait()

    # Check
    if ( 0 != return_code ):
        raise Exception( p.stderr.read().decode( 'utf-8' )[:-1] )

    return p.stdout.read().decode( 'utf-8' )[:-1]



def execute_and_stream_output( cmd, out_stream ):
    '''
    Invoke and execute a subprocess.
    Output from the process to stdout is sent line by line to the specified output stream.

    :param cmd: sequence of program arguments or else a single string or path-like object (see Python "subprocess" module).
    :param out_stream: output stream
    :return: None
    '''
    p = subprocess.Popen(
        cmd,
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
    )

    # Read output line by line and display via output stream.
    while True:
        line = p.stdout.readline()
        if not line: break
        out_stream( line.decode( 'utf-8' )[:-1] )

    # Wait for process to terminate and retrieve return code.
    return_code = p.wait()

    if ( 0 != return_code ):
        raise Exception( p.stderr.read().decode( 'utf-8' )[:-1] )


from ..util.config_data import ConfigData
from ..util.execute import execute


def clear_sim( setup_dir, id ):
    '''
    Delete containers of finished simulations.

    :param setup_dir: path to simulation setup (string)
    :param id: either 'all' or ID of simulation container to be cleared (string)
    :return: on success, return list of cleared simulation IDs (list of string)
    '''

    if not isinstance( id, str ):
        raise TypeError( 'Parameter \'id\' must be of type \'str\'' )

    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )
    sim_ids_down = config_data['sim_ids_down']

    if 'all' == id.lower():
        rm_ids = list( sim_ids_down )
    else:
        rm_ids = [ id ]
        if id not in sim_ids_down:
            raise RuntimeError( 'No finished simulation (status \'DOWN\') with ID \'{}\''.format( id ) )

    if 0 != len( rm_ids ):
        execute( [
            'docker', 'rm', # Remove container.
            '--volumes', # Remove anonymous volumes associated with the container.
            *rm_ids # Simulation ID (Docker container name).
        ] )

    if 'all' == id.lower():
        sim_ids_down.clear()
    else:
        del sim_ids_down[ sim_ids_down.index( *rm_ids ) ]

    # Save simulation setup configuration.
    config_data.write()

    # On success, return cleared simulation ID.
    return rm_ids


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Delete containers of finished simulations.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    group = parser.add_mutually_exclusive_group( required = True )

    group.add_argument(
        '--id', 
        action = 'store',
        metavar = 'ID',
        help = 'simulation ID (Docker container name)'
    )

    group.add_argument(
        '--all',
        action = 'store_true',
        help = 'remove all simulation containers'
    )

    args = parser.parse_args()

    try:
        id = 'all' if ( True == args.all ) else args.id
        sim_ids = clear_sim( args.setup_dir, id )

        print( 'Cleared simulation(s) with ID = {}'.format( ' '.join( sim_ids ) ) )
        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



import pathlib

from .check_sim_setup import check_sim_setup
from ..util.config_data import ConfigData


def configure_sim_setup( 
        setup_dir,
        docker_file,
        scenario_file,
        extra_files = [],
        extra_dirs = [],
        results = []
        ):
    '''
    Configure an existing simulation setup.

    :param setup_dir: directory of the simulation setup (string, default: '.')
    :param docker_file: name of the Dockerfile used for building the simulation orchestrator image (string)
    :param scenario_file: name of the mosaik scenario file (string)
    :param extra_files: additional files to be added to the simulation orchestrator image (list of strings)
    :param extra_dirs: additional directories to be added to the simulation orchestrator image (list of strings)
    :param results: list of paths of result files or folders, i.e., files or folders produced by the simulation that should be retrieved after the simulation has finished (list of strings)
    :return: on success, return absolute path to simulation setup config file (string)
    '''

    if not isinstance( setup_dir, str ):
        raise TypeError( 'Parameter \'setup_dir\' must be of type \'str\'' )

    if not isinstance( docker_file, str ):
        raise TypeError( 'Parameter \'docker_file\' must be of type \'str\'' )

    if not isinstance( scenario_file, str ):
        raise TypeError( 'Parameter \'scenario_file\' must be of type \'str\'' )

    if not isinstance( extra_files, list ) or not all( isinstance( elem, str ) for elem in extra_files ):
        raise TypeError( 'Parameter \'extra_files\' must be of type \'list of str\'' )

    if not isinstance( extra_dirs, list ) or not all( isinstance( elem, str ) for elem in extra_dirs ):
        raise TypeError( 'Parameter \'extra_dirs\' must be of type \'list of str\'' )

    if not isinstance( results, list ) or not all( isinstance( elem, str ) for elem in results ):
        raise TypeError( 'Parameter \'results\' must be of type \'list of str\'' )

    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )

    config_data_orch = config_data['orchestrator']
    config_data_orch['docker_file'] = docker_file
    config_data_orch['scenario_file'] = scenario_file
    config_data_orch['extra_files'] = extra_files
    config_data_orch['extra_dirs'] = extra_dirs
    config_data_orch['results'] = results

    # Save simulation setup configuration.
    config_data.write()

    # Check if the new configuration is valid.
    check_setup = check_sim_setup( setup_dir )
    if not check_setup['valid']:
        raise RuntimeError( check_setup['status'] )

    # On success, return absolute path to simulation setup config file.
    return str( config_data.path )


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Configure an existing simulation setup.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    required = parser.add_argument_group( 'required named arguments' )
    
    required.add_argument(
        '-d', '--docker-file', 
        action = 'store',
        required = True,
        metavar = 'DOCKER_FILE',
        help = 'name of Dockerfile for orchestrator'
    )

    required.add_argument(
        '-s', '--scenario-file', 
        action = 'store',
        required = True,
        metavar = 'SCENARIO_FILE',
        help = 'name of main mosaik script for orchestrator'
    )

    parser.add_argument(
        '--extra-file', 
        action = 'append',
        metavar = 'EXTRA_FILE',
        help = 'additional file to be added to the orchestrator Docker image'
    )

    parser.add_argument(
        '--extra-dir', 
        action = 'append',
        metavar = 'EXTRA_DIR',
        help = 'additional directory to be added to the orchestrator Docker image'
    )

    parser.add_argument(
        '--result', 
        action = 'append',
        metavar = 'RESULT',
        help = 'paths of result file or folder, i.e., file or folder produced by the simulation that should be retrieved after the simulation has finished (list of strings)'
    )

    args = parser.parse_args()

    try:
    
        sim_config_file = configure_sim_setup( 
            args.setup_dir, 
            args.docker_file,
            args.scenario_file,
            args.extra_file,
            args.extra_dir,
            args.result
        )

        print( 'Updated simulation configuration file: {}'.format( sim_config_file ) )
        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



from .._config import ORCH_IMAGE_NAME_TEMPLATE
from ..util.config_data import ConfigData
from ..util.create_unique_id import create_unique_id
from ..util.execute import execute


def start_sim( setup_dir, id = None ):
    '''
    Start a new simulation.

    :param setup_dir: path to simulation setup (string)
    :param id: ID of new simulation (string, default: None)
    :return: on success, return new simulation ID (int)
    '''

    if not id == None and not isinstance( id, str ):
        raise TypeError( 'Parameter \'id\' must be of type \'str\'' )

    if id == None:
        id = create_unique_id()

    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )

    sim_setup_id = config_data['id'].strip()
    config_data_orch = config_data['orchestrator']
    scenario_file = config_data_orch['scenario_file'].strip()

    sim_ids_up = config_data['sim_ids_up']
    sim_ids_down = config_data['sim_ids_down']

    if id in sim_ids_up or id in sim_ids_down:
        raise RuntimeError( 'Simulation ID \'{}\' has already been used'.format( id ) )

    # Define Docker image name.
    docker_image_name = ORCH_IMAGE_NAME_TEMPLATE.format( sim_setup_id.lower() )

    execute( [
        'docker', 'run', # Docker run command.
        '--detach', # Run container in background.
        #'--rm', # '-it', # Only for debugging.
        '--name', id, # Specify container name as simulation id.
        '--env', 'SCENARIO_FILE={}'.format( scenario_file ), # Specify scenario file.
        docker_image_name # Specify the Docker image.
    ] )

    # Update sim setup config.
    sim_ids_up.append( id )

    # Save simulation setup configuration.
    config_data.write()

    # On success, return new simulation ID.
    return id


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Start a new simulation.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    parser.add_argument(
        'id',
        nargs = '?',
        default = None,
        metavar = 'ID',
        help = 'simulation ID (Docker container name)'
    )

    args = parser.parse_args()

    try:
        sim_id = start_sim( args.setup_dir, args.id )

        print( 'Started new simulation with ID = {}'.format( sim_id ) )
        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



from .._config import DOCKER_STOP_WAIT_TIME
from ..util.config_data import ConfigData
from ..util.execute import execute


def cancel_sim( setup_dir, id ):
    '''
    Cancel a simulation (stop simulation container).

    :param setup_dir: path to simulation setup (string)
    :param id: either 'all' or ID of running simulation container (string)
    :return: on success, return ID of cancelled simulation (int)
    '''

    if not id == None and not isinstance( id, str ):
        raise TypeError( 'Parameter \'id\' must be of type \'str\'' )

    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )

    sim_ids_up = config_data['sim_ids_up']
    sim_ids_down = config_data['sim_ids_down']

    if 'all' == id.lower():
        stop_ids = list( sim_ids_up ) # Deep copy of list!
    else:
        stop_ids = [ id ]

        if not id in sim_ids_up:
            raise RuntimeError( 'No running simulation (status \'UP\') with ID \'{}\''.format( id ) )

        if id in sim_ids_down:
            raise RuntimeError( 'There is already a finished simulation (status \'DOWN\') with ID \'{}\''.format( id ) )

    if 0 != len( stop_ids ):
        execute( [
            'docker', 'stop', # Stop Docker container.
            '--time' , str( DOCKER_STOP_WAIT_TIME ), # Seconds to wait for stop before killing it.
            *stop_ids
        ] )

    # Update sim setup config.
    if 'all' == id.lower():
        sim_ids_up.clear()
        sim_ids_down.extend( stop_ids )
    else:
        del sim_ids_up[ sim_ids_up.index( *stop_ids ) ]
        sim_ids_down.append( *stop_ids )

    # Save simulation setup configuration.
    config_data.write()

    # On success, return cancelled simulation ID.
    return stop_ids


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Cancel a simulation (stop simulation container).'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    group = parser.add_mutually_exclusive_group( required = True )

    group.add_argument(
        '--id',
        action = 'store',
        metavar = 'ID',
        help = 'simulation ID (Docker container name)'
    )

    group.add_argument(
        '--all',
        action = 'store_true',
        help = 'cancel all running simulations'
    )

    args = parser.parse_args()

    try:
        id = 'all' if ( True == args.all ) else args.id
        sim_id = cancel_sim( args.setup_dir, id )

        print( 'Cancelled simulation(s) with ID = {}'.format( ' '.join( sim_id ) ) )
        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



from ..util.config_data import ConfigData


def get_sim_ids( setup_dir ):
    '''
    Get IDs all running (status 'UP') and finished (status 'DOWN') simulations of a simulation setup.

    :param setup_dir: path to simulation setup (string)
    :return: return dict with simulation IDs in the following format:
        {
            'up': [string] # IDs of running simulations
            'down': [string] # IDs of finished simulations
        }
    '''
    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )

    sim_ids_up = config_data['sim_ids_up']
    sim_ids_down = config_data['sim_ids_down']

    status = dict( up = sim_ids_up, down = sim_ids_down )

    return status


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Get IDs all running (status \'UP\') and finished (status \'DOWN\') simulations of a simulation setup.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    args = parser.parse_args()

    try:

        sim_ids = get_sim_ids( args.setup_dir )

        sim_ids_up = sim_ids['up']
        sim_ids_down = sim_ids['down']

        msg = 'running:\n' + '\t{}\n'*len( sim_ids_up ) + 'finished:\n' + '\t{}\n'*len( sim_ids_down)
        print( msg.format( *sim_ids_up, *sim_ids_down ) )

        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



import pathlib

from ..util.config_data import ConfigData


def get_sim_setup_root( dir ):
    '''
    Check if the specified directory (or any parent directory) contains a simulation setup configuration.

    :param dir: directory path to check (string)
    :return: return the following dict:
        {
            'valid': boolean # flag indicating if this directory (or any parent directory) contains a simulation setup configuration
            'dir': string # directory containing a simulation setup configuration if 'valid', otherwise empty
        }
    '''
    # Resolve the path.
    dir_path = pathlib.Path( dir ).resolve( strict = True )

    if not dir_path.is_dir():
        raise RuntimeError( 'not a directory path: {}'.format( dir_path ) )

    # Iterate through all parent directories and check if they contain a simulation setup configuration.
    for p in [ dir_path, *dir_path.parents ]:
        try:
            # Retrieve simulation setup configuration.
            config_data = ConfigData( p )

            return dict( valid = True, dir = str( p ) )
        except: # Exception as err:
            #print( err )
            pass

    # Nothing found.
    return dict( valid = False, dir = '' )


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Check if the specified directory (or any parent directory) contains a simulation setup configuration.'
    )

    parser.add_argument(
        'dir',
        nargs = '?',
        default = '.',
        metavar = 'DIR',
        help = 'folder path (default: current working directory)'
    )

    args = parser.parse_args()

    try:
        setup_root = get_sim_setup_root( args.dir )

        if True == setup_root['valid']:
            print( setup_root['dir'] )
            sys.exit( 0 )
        else:
            print( 'This directory is not part of a simulation setup: {}'.format( pathlib.Path( args.dir ).resolve() ) )
            sys.exit( 1 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )


import pathlib

from ..util.config_data import ConfigData


def check_sim_setup( setup_dir ):
    '''
    Check if simulation setup is valid.

    :param setup_dir: path to simulation setup (string)
    :return: return dict with status of setup check in the following format:
        {
            'valid': boolean # flag indicating if setup is valid
            'status': string # detailed status message
        }
    '''

    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )

    err = '"{}" is missing in configuration file'

    if not 'id' in config_data:
        return dict( valid = False, status = err.format( 'id' ) )
    elif not 'orchestrator' in config_data:
        return dict( valid = False, status = err.format( 'orchestrator' ) )
    elif not 'sim_ids_up' in config_data:
        return dict( valid = False, status = err.format( 'sim_ids_up' ) )
    elif not 'sim_ids_down' in config_data:
        return dict( valid = False, status = err.format( 'sim_ids_down' ) )

    config_data_orch = config_data['orchestrator']
    if not 'scenario_file' in config_data_orch:
        return dict( valid = False, status = err.format( 'orchestrator.scenario_file' ) )
    elif not 'docker_file' in config_data_orch:
        return dict( valid = False, status = err.format( 'orchestrator.docker_file' ) )
    elif not 'extra_files' in config_data_orch:
        return dict( valid = False, status = err.format( 'orchestrator.extra_files' ) )
    elif not 'extra_dirs' in config_data_orch:
        return dict( valid = False, status = err.format( 'orchestrator.extra_dirs' ) )
    elif not 'results' in config_data_orch:
        return dict( valid = False, status = err.format( 'orchestrator.results' ) )

    try:
        pathlib.Path( setup_dir, config_data_orch['scenario_file'] ).resolve( strict = True )
    except Exception as err:
        return dict( valid = False, status = 'scenario file missing\n{}'.format( err ) )
        
    try:
        pathlib.Path( setup_dir, config_data_orch['docker_file'] ).resolve( strict = True )
    except Exception as err:
        return dict( valid = False, status = 'Dockerfile missing\n{}'.format( err ) )

    try:
        for file in config_data_orch['extra_files']:
            if not pathlib.Path( setup_dir, file ).resolve( strict = True ).is_file():
                raise Exception( 'not a file: {}'.format( file ) )
    except Exception as err:
        return dict( valid = False, status = 'extra file not found\n{}'.format( err ) )

    try:
        for dir in config_data_orch['extra_dirs']:
            if not pathlib.Path( setup_dir, dir ).resolve( strict = True ).is_dir():
                raise Exception( 'not a directory: {}'.format( dir ) )
    except Exception as err:
        return dict( valid = False, status = 'extra directory not found\n{}'.format( err ) )

    return dict( valid = True, status = 'simulation setup is valid: {}'.format( setup_dir ) )


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Check if simulation setup is valid.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    args = parser.parse_args()

    try:
        setup_status = check_sim_setup( args.setup_dir )

        print( setup_status['status'] )
        if True == setup_status['valid']:
            sys.exit( 0 )
        else:
            sys.exit( 1 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )


def flatten_list( lst ):
	'''
    Flatten a list of lists, i.e., transform
        [ [ a1, ..., aN ], [ b1, ..., bN ], ... ]
    into
        [ a1, ..., aN, b1, ..., bN, ... ]
        
    :param lst: list of lists
    :return: flattened list
	'''
	return [item for sublist in lst for item in sublist]


from .starter import start_dockerimage, start_dockerfile

from mosaik.simmanager import StarterCollection

# Add new simulation starters.
s = StarterCollection()
s['dockerimage'] = start_dockerimage
s['dockerfile'] = start_dockerfile


from setuptools import setup, find_packages


# Read long description from file (reStructuredText syntax). Will be parsed and displayed as HTML online.
with open( 'description.rst' ) as description_file:
  _long_description = description_file.read()

setup(
    name = 'mosaik-docker',
    maintainer = 'ERIGrid 2.0 development team',
    maintainer_email = 'edmund.widl@ait.ac.at',
    url = 'https://mosaik-docker.readthedocs.io',
    version = '0.1.2',
    platforms = [ 'any' ],
    packages = find_packages(),
    package_data = {
        'mosaik_docker': [
            'sim_setup_template/cookiecutter.json',
            'sim_setup_template/{{cookiecutter.directory_name}}/mosaik-docker.json',
            'sim_setup_template/{{cookiecutter.directory_name}}/README.md',
            'sim_setup_template/{{cookiecutter.directory_name}}/dockerfiles/*',
        ],
    },
    install_requires = [
        'cookiecutter',
        'mosaik'
    ],
    entry_points={
        'console_scripts': [
            'create_sim_setup = mosaik_docker.cli.create_sim_setup:main',
            'configure_sim_setup = mosaik_docker.cli.configure_sim_setup:main',
            'check_sim_setup = mosaik_docker.cli.check_sim_setup:main',
            'build_sim_setup = mosaik_docker.cli.build_sim_setup:main',
            'get_sim_setup_root = mosaik_docker.cli.get_sim_setup_root:main',
            'start_sim = mosaik_docker.cli.start_sim:main',
            'cancel_sim = mosaik_docker.cli.cancel_sim:main',
            'clear_sim = mosaik_docker.cli.clear_sim:main',
            'get_sim_status = mosaik_docker.cli.get_sim_status:main',
            'get_sim_results = mosaik_docker.cli.get_sim_results:main',
            'get_sim_ids = mosaik_docker.cli.get_sim_ids:main',
            'delete_sim_setup = mosaik_docker.cli.delete_sim_setup:main',
        ]
    },
    description = 'This package eases the deployment of the mosaik co-simulation framework with Docker.',
    long_description = _long_description,
    license = 'BSD 3-Clause License',
    keywords = [
        'mosaik',
        'Docker',
    ],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering'
    ],
)



from cookiecutter.main import cookiecutter
import pathlib

from ..util.create_unique_id import create_unique_id


def create_sim_setup( name, dir = '.', id = None ):
    '''
    Create an empty simulation setup in a new directory.

    :param name: name of the simulation setup (string)
    :param dir: directory to put the generated simulation setup (string, default: '.')
    :param id: unique ID for the simulation setup (string, default: None)
    :return: on success, return absolute path to created simulation setup directory (string)
    '''

    if not isinstance( name, str ):
        raise TypeError( 'Parameter \'name\' must be of type \'str\'' )

    if not isinstance( dir, str ):
        raise TypeError( 'Parameter \'dir\' must be of type \'str\'' )

    try:
        dir_path = pathlib.Path( dir ).resolve( strict = True )
    except Exception as err:
        raise FileNotFoundError( '\'{}\' is not a valid directory name'.format( dir ) )

    if pathlib.Path( dir_path, name ).is_dir():
        raise FileExistsError( 'Directory \'{}\' already exists'.format( pathlib.Path( dir_path, name ).resolve( strict = False ) ) )

    if not id == None and not isinstance( id, str ):
        raise TypeError( 'Parameter \'id\' must be of type \'str\'' )

    if id == None:
        id = create_unique_id()

    # Path to cookiecutter template directory.
    template_path = pathlib.Path( 
        pathlib.Path( __file__ ).parent,
        '..', 
        'sim_setup_template' 
        ).resolve( strict = True )

    # Create dict with template parameters.
    template_params = {
        'directory_name': name,
        'id': id
        }

    # Create project from the cookiecutter template.
    cookiecutter(
        str( template_path ),
        output_dir = dir,
        no_input = True,
        extra_context = template_params
        )

    # On success, return name of created simulation setup directory.
    return pathlib.Path( dir_path, name ).resolve()


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Create an empty simulation setup in a new directory.'
    )

    parser.add_argument(
        'name', 
        metavar = 'NAME',
        help = 'name of the new simulation setup'
    )

    parser.add_argument(
        'dir',
        nargs = '?',
        default = '.',
        metavar = 'DIR',
        help = 'directory to put the generated simulation setup (default: current working directory)'
    )

    parser.add_argument(
        '--id', 
        action = 'store',
        default = None,
        metavar = 'ID',
        help = 'unique ID for the new simulation setup'
    )

    args = parser.parse_args()

    try:
        sim_setup_dir = create_sim_setup( args.name, args.dir, args.id )

        print( 'Created new simulation setup: {}'.format( sim_setup_dir ) )
        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



import json
import pathlib

from .._config import CONFIG_FILE_NAME


class ConfigData:
    '''
    This class handles access to simulation setup configuration data.
    '''

    # Constructor.
    def __init__( self, setup_dir ):

        if not ( isinstance( setup_dir, str ) or isinstance( setup_dir, pathlib.Path ) ):
            raise TypeError( 'Parameter \'setup_dir\' must be of type \'str\' or \'pathlib.Path\'' )

        try:
            setup_dir_path = pathlib.Path( setup_dir ).resolve( strict = True )
        except Exception as err:
            raise RuntimeError( 'not a valid directory: {}\n{}'.format( setup_dir, err ) )

        # Load sim setup configuration.
        try:
            self.__sim_setup_file_path = pathlib.Path( setup_dir_path, CONFIG_FILE_NAME ).resolve( strict = True )
        except Exception as err:
            raise RuntimeError( 'not a valid simulation setup: {}\n{}'.format( setup_dir_path, err ) )

        with open( self.__sim_setup_file_path ) as sim_setup_file:
            try:
                self.__config_data = json.load( sim_setup_file )
            except Exception as err:
                raise Exception( 'Invalid JSON format: {}\n{}'.format( self.__sim_setup_file_path, str( err ) ) )


    def __setitem__( self, index, value ):
        '''
        For setting a configuration value.
        '''
        self.__config_data[index] = value


    def __getitem__( self, index ):
        '''
        For retrieving a configuration value.
        '''
        return self.__config_data[index]


    def __contains__( self, item ):
        '''
        Returns a boolean value depending on whether the configuration contains the specified item or not.
        '''
        return item in self.__config_data


    def write( self ):
        '''
        Save configuration.
        '''
        with open( self.path, 'w' ) as sim_setup_file:
            json.dump(
                self.__config_data,
                sim_setup_file,
                indent = 2,
                separators = ( ',', ': ' ) )
            sim_setup_file.write( '\n' )


    @property
    def path( self ):
        '''
        Absolute path to configuration file.
        '''
        return self.__sim_setup_file_path


    @property
    def data( self ):
        '''
        Configuration data as dict.
        '''
        return self.__config_data



import pathlib
import shutil

from .._config import ORCH_CONTEXT_DIR_NAME, ORCH_CONTEXT_EXTRA_DIR_NAME, ORCH_IMAGE_NAME_TEMPLATE
from ..util.execute import execute_and_stream_output
from ..util.config_data import ConfigData


def build_sim_setup( setup_dir, out_stream = print ):
    '''
    Build simulation setup as preparation for running the simulation.
    This includes building the Docker image of the mosaik orchestrator.

    :param setup_dir: path to simulation setup (string)
    :param out_stream: output from the build process to stderr will be piped to this stream (callable)
    :return: return dict with status of build process:
        {
            'valid': flag indicating if build succeded (boolean)
            'status': detailed status message (string)
        }
    '''

    if not callable( out_stream ):
        raise TypeError( 'Parameter \'out_stream\' must be callable' )

    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )

    try:
        # Retrieve data from config file.
        sim_setup_id = config_data['id'].strip()
        config_data_orch = config_data['orchestrator']
        scenario_file = config_data_orch['scenario_file'].strip()
        docker_file = config_data_orch['docker_file'].strip()
        extra_files = [ f.strip() for f in config_data_orch['extra_files'] ]
        extra_dirs = [ d.strip() for d in config_data_orch['extra_dirs'] ]

        # Check if scenario file and Dockerfile exists.
        scenario_file_path = pathlib.Path( setup_dir, scenario_file ).resolve( strict = True )
        docker_file_path = pathlib.Path( setup_dir, docker_file ).resolve( strict = True )
        extra_file_paths = [ pathlib.Path( setup_dir, f ).resolve( strict = True ) for f in extra_files ]
        extra_dir_paths = [ pathlib.Path( setup_dir, d ).resolve( strict = True ) for d in extra_dirs ]

        # Define path for "context directory". All resources needed to create the orchestrator image will be copied there.
        orch_context_dir = pathlib.Path( setup_dir, ORCH_CONTEXT_DIR_NAME ).resolve( strict = False )
        orch_context_extra_dir = pathlib.Path( orch_context_dir, ORCH_CONTEXT_EXTRA_DIR_NAME ).resolve( strict = False )

        # Check if context directory already exists. If yes, delete it.
        if orch_context_dir.is_dir():
            shutil.rmtree( orch_context_dir )

        # Create new context directory.
        orch_context_dir.mkdir( exist_ok = False )
        orch_context_extra_dir.mkdir( exist_ok = False )

        # Copy resources to context directory.
        shutil.copy( scenario_file_path, orch_context_dir )
        for f in extra_file_paths:
            shutil.copy( f, orch_context_extra_dir )
        for d in extra_dir_paths:
            shutil.copytree( d, pathlib.Path( orch_context_extra_dir, d.name ) )

        # Define Docker image name.
        docker_image_name = ORCH_IMAGE_NAME_TEMPLATE.format( sim_setup_id.lower() )

        cmd = [
            'docker', 'build', # Docker build command.
            '-t', docker_image_name, # Specify image name.
            '--build-arg', 'SCENARIO_FILE={}'.format( scenario_file ), # Specify scenario file.
            '--build-arg', 'EXTRA={}'.format( ORCH_CONTEXT_EXTRA_DIR_NAME ), # Specify directory with extra files and directories.
            '-f', docker_file_path, # Specify the Dockerfile.
            orch_context_dir # Specify the build context.
        ]
        
        execute_and_stream_output( cmd, out_stream )

    except Exception as err:
        return dict( 
            valid = False, 
            status = 'building simulation setup failed:\n{}\nrun "check_sim_setup" for details'.format( err )
        )

    return dict( 
        valid = True, 
        status = 'building simulation setup succeeded: {}'.format( config_data.path.parent ) 
    )


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Build simulation setup as preparation for running the simulation. This includes building the Docker image of the mosaik orchestrator.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    args = parser.parse_args()

    try:
        build_status = build_sim_setup( args.setup_dir )

        print( build_status['status'] )
        if True == build_status['valid']:
            sys.exit( 0 )
        else:
            sys.exit( 1 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )


from ..util.config_data import ConfigData
from ..util.execute import execute_and_capture_output
from ..util.flatten_list import flatten_list


def get_sim_status( setup_dir ):
    '''
    Get status of all simulations of a mosaik-docker setup.
    Updates the simulation setup information about which containers are running (status UP) or finished (status DOWN) if it is not up to date.

    :param setup_dir: path to simulation setup (string)
    :return: return dict with status information for all simulations of a mosaik-docker simulation setup in the following format:
        {
            'up': { string: string } # running simulation IDs and status
            'down': { string: string } # finished simulation IDs and status
        }
    '''

    # Retrieve simulation setup configuration.
    config_data = ConfigData( setup_dir )
    sim_ids_up = config_data['sim_ids_up']
    sim_ids_down = config_data['sim_ids_down']

    # Check if any simulation with status 'running' has already finished.
    sim_ids_stopped = _diff_status( sim_ids_up, 'running' )
    if 0 != len( sim_ids_stopped ):
        # Update sim setup config accordingly.
        for id in sim_ids_stopped:
            del sim_ids_up[ sim_ids_up.index( id ) ]
            sim_ids_down.append( id )

        # Save simulation setup configuration.
        config_data.write()

    # Retrieve status information from Docker.
    sim_status_up = _retrieve_status( sim_ids_up )
    sim_status_down = _retrieve_status( sim_ids_down )

    return dict(
        up = sim_status_up,
        down = sim_status_down
        )


def _diff_status( ids, status ):
    '''
    Check if Docker containers are in a specified status.
    Returns list of Docker container names that are not in the specified status.

    :param ids: list simulation IDs (Docker container names)
    :param status: supposed status of Docker containers
    :return: list of container names that are not in the supposed status
    '''
    # Define filter for simulations IDs (container names).
    id_filter = [ [ '--filter', 'name={}'.format( id ) ] for id in ids ]

    # Define filter for status.
    status_filter = [ '--filter', 'status={}'.format( status ) ]

    # Retrieve container names (as string).
    out = execute_and_capture_output( [
        'docker', 'ps', # List containers.
        '--no-trunc', # Do not truncate output.
        '--all', # Show all containers (default shows just running).
        *flatten_list( id_filter ), # Filter output based on conditions provided.
        *status_filter, # Filter output based on conditions provided.
        '--format', '{{.Names}}' # Only output container names.
    ] )

    # Parse string to list.
    confirmed_status = out.split( '\n' )

    # Return list of container names that are not in the specified status.
    status_diff = [ id for id in ids if id not in confirmed_status ]
    return status_diff


def _retrieve_status( ids ):
    '''
    Retrieve status information of Docker containers.

    :param ids: list of simulation IDs (Docker container names)
    :return: dict of container status information
    '''
    if 0 == len( ids ):
        return {}

    # Define filter for IDs.
    id_filter = [ [ '--filter', 'name={}'.format( id ) ] for id in ids ]

    # Retrieve container status (as string).
    out = execute_and_capture_output( [
        'docker', 'ps', # List containers.
        '--no-trunc', # Do not truncate output.
        '--all', # Show all containers (default shows just running).
        *flatten_list( id_filter ), # Filter output based on conditions provided.
        '--format', '{{.Names}}:{{.Status}}' # Only output conatainer name and status.
    ] )

    # Parse string to dict.
    status_dict = { s[0]: s[1] for s in [ o.split( ':', 1 ) for o in out.split( '\n' ) ] }

    return status_dict


def main():

    import argparse
    import sys

    # Command line parser.
    parser = argparse.ArgumentParser(
        description = 'Get status of all simulations of a mosaik-docker setup. Updates the simulation setup information about which containers are running (status UP) or finished (status DOWN) if it is not up to date.'
    )

    parser.add_argument(
        'setup_dir',
        nargs = '?',
        default = '.',
        metavar = 'SETUP_DIR',
        help = 'path to simulation setup directory (default: current working directory)'
    )

    args = parser.parse_args()

    try:

        status = get_sim_status( args.setup_dir )

        sims_up = status['up']
        sims_down = status['down']

        msg = 'running:\n' + '\t{}\n'*len( sims_up ) + 'finished:\n' + '\t{}\n'*len( sims_down )
        str_sims_up = [ '{}: {}'.format( x[0], x[1] ) for x in sims_up.items() ]
        str_sims_down = [ '{}: {}'.format( x[0], x[1] ) for x in sims_down.items() ]
        print( msg.format( *str_sims_up, *str_sims_down ) )

        sys.exit( 0 )

    except Exception as err:

        print( str( err ) )
        sys.exit( 3 )



import uuid


def create_unique_id( n_digits = 6 ):
    '''
    Create a unique ID.

    :param n_digits: number of digits of the returned ID (int, default: 6)

    :return: unique alphanumeric ID (string)
    '''

    return str( uuid.uuid4() ).replace( '-', '' )[:n_digits]



from mosaik.simmanager import make_proxy
from mosaik.exceptions import ScenarioError

import os
import sys
import shlex
import subprocess



def start_dockerimage(world, sim_name, sim_config, sim_id, sim_params):
    """
    ...
    """
    print('###### start_dockerimage ######')
    replacements = {
        'addr': '%s:%s' % (world.config['addr'][0], world.config['addr'][1]),
        'python': sys.executable,
    }
    cmd = sim_config['dockerimage'] % replacements
    if 'posix' in sim_params.keys():
        posix = sim_params.pop('posix')
        cmd = shlex.split(cmd, posix=posix)
    else:
        cmd = shlex.split(cmd, posix=(os.name != 'nt'))
    cwd = sim_config['cwd'] if 'cwd' in sim_config else '.'

    # Make a copy of the current env. vars dictionary and update it with the
    # user provided values (or an empty dict as a default):
    env = dict(os.environ)
    env.update(sim_config.get('env', {}))

    kwargs = {
        'bufsize': 1,
        'cwd': cwd,
        'universal_newlines': True,
        'env': env,  # pass the new env dict to the sub process
    }
    try:
        proc = subprocess.Popen(cmd, **kwargs)
    except (FileNotFoundError, NotADirectoryError) as e:
        # This distinction has to be made due to a change in python 3.8.0.
        # It might become unecessary for future releases supporting
        # python >= 3.8 only.
        if str(e).count(':')==2:
            eout = e.args[1]
        else:
            eout = str(e).split('] ')[1]
        raise ScenarioError('Simulator "%s" could not be started: %s'
                            % (sim_name, eout)) from None

    proxy = make_proxy(world, sim_name, sim_config, sim_id, sim_params,
                       proc=proc)
    return proxy
	

def start_dockerfile(world, sim_name, sim_config, sim_id, sim_params):
    """
    ...
    """
    print('###### start_dockerfile ######')
    addr = sim_config['connect']
    try:
        host, port = addr.strip().split(':')
        addr = (host, int(port))
    except ValueError:
        raise ScenarioError('Simulator "%s" could not be started: Could not '
                            'parse address "%s"' %
                            (sim_name, sim_config['connect'])) from None

    proxy = make_proxy(world, sim_name, sim_config, sim_id, sim_params,
                       addr=addr)
    return proxy